"""Provides YAML parsing support."""

from typing import IO, Any, Dict, Mapping, Optional

import yaml
from django.conf import settings
from django.utils.encoding import force_str
from rest_framework.exceptions import ParseError
from rest_framework.parsers import BaseParser


class YAMLParser(BaseParser):
    """Parse YAML-serialized data."""

    media_type = "application/yaml"

    # ruff: noqa: ARG002
    def parse(
        self,
        stream: IO[Any],
        media_type: Optional[str] = None,
        parser_context: Optional[Mapping[str, Any]] = None,
    ) -> Dict[str, Any]:
        """Parse the incoming bytestream as YAML and returns the resulting data."""
        parser_context = parser_context or {}
        encoding = parser_context.get("encoding", settings.DEFAULT_CHARSET)

        try:
            data = stream.read().decode(encoding)
            parsed_data = yaml.safe_load(data)

        except (ValueError, yaml.parser.ParserError) as exc:
            raise ParseError("YAML parse error - %s" % force_str(exc)) from exc

        if not isinstance(parsed_data, dict):
            raise ParseError("Expected dict, got %s" % type(parsed_data))

        return parsed_data



"""Provides YAML rendering support."""

from typing import Any, Mapping, Optional

import yaml
from rest_framework.renderers import BaseRenderer

from .encoders import SafeDumper


class YAMLRenderer(BaseRenderer):
    """Renderer which serializes to YAML."""

    media_type = "application/yaml"
    format = "yaml"
    encoder = SafeDumper
    charset = "utf-8"
    ensure_ascii = False
    default_flow_style = False
    sort_keys = False

    def render(
        self,
        data: Any,
        _accepted_media_type: Optional[str] = None,
        _renderer_context: Optional[Mapping[str, Any]] = None,
    ) -> bytes:
        """Render `data` into serialized YAML."""
        if data is None:
            return b""

        return yaml.dump(  # type: ignore [no-any-return]
            data,
            sort_keys=self.sort_keys,
            stream=None,
            encoding=self.charset,
            Dumper=self.encoder,
            allow_unicode=not self.ensure_ascii,
            default_flow_style=self.default_flow_style,
        )



"""
Custom DRF fields for use in serialization when using the YAMLRenderer.

These fields are used to render the YAML in a more readable way.
"""
from typing import Any

from rest_framework import fields

from . import styles


class FlowStyleDictField(fields.DictField):
    """A DRF field which renders as a YAML flow style mapping."""

    def to_representation(self, value: Any) -> dict[Any, Any]:
        """Render a flow style mapping."""
        return styles.flow_style_mapping(super().to_representation(value))


class FlowStyleListField(fields.ListField):
    """A DRF field which renders as a YAML flow style sequence."""

    def to_representation(self, value: Any) -> list[Any]:
        """Render a flow style sequence."""
        return styles.flow_style_sequence(super().to_representation(value))


class SingleQuotedCharField(fields.CharField):
    """A DRF field which renders as a YAML single quoted string."""

    def to_representation(self, value: Any) -> str:
        """Render a single quoted string."""
        return styles.single_quoted_str(super().to_representation(value))


class DoubleQuotedCharField(fields.CharField):
    """A DRF field which renders as a YAML double quoted string."""

    def to_representation(self, value: Any) -> str:
        """Render a double quoted string."""
        return styles.double_quoted_str(super().to_representation(value))


class FoldedCharField(fields.CharField):
    """A DRF field which renders as a YAML folded string."""

    def to_representation(self, value: Any) -> str:
        """Render a folded string."""
        return styles.folded_str(super().to_representation(value))


class LiteralCharField(fields.CharField):
    """A DRF field which renders as a YAML literal string."""

    def to_representation(self, value: Any) -> str:
        """Render a literal string."""
        return styles.literal_str(super().to_representation(value))



"""String types for YAML."""


class FoldedStr(str):
    """A string which should be rendered as a YAML folded string."""


class LiteralStr(str):
    """A string which should be rendered as a YAML literal string."""


class SingleQuotedStr(str):
    """A string which should be rendered as a YAML single quoted string."""


class DoubleQuotedStr(str):
    """A string which should be rendered as a YAML double quoted string."""


class FlowStyleSequence(list):
    """A sequence which should be rendered as a YAML flow style sequence."""


class FlowStyleMapping(dict):
    """A mapping which should be rendered as a YAML flow style mapping."""



"""Helper classes for parsers."""
import decimal
import types
from collections import OrderedDict
from datetime import time, timedelta
from uuid import UUID

import yaml
from django.utils.encoding import force_str
from django.utils.safestring import SafeString
from rest_framework.exceptions import ErrorDetail
from rest_framework.relations import Hyperlink
from rest_framework.utils.serializer_helpers import ReturnDict, ReturnList

from . import styles


class SafeDumper(yaml.SafeDumper):
    """
    Dumper subclass that handles commonly used DRF types.

    These are the supported types:
    - Decimal is represented as a string.
    - UUID is represented as a string.
    - Time is represented as a string.
    - TimeDelta is represented as a string.
    - Hyperlink is represented as a string.
    - ErrorDetail is represented as a string.
    - SafeString is represented as a string.
    - OrderedDict is represented as a dict.
    - ReturnDict is represented as a dict.
    - ReturnList is represented as a list.
    - Generators are represented as a list.

    It also supports the following custom types:
    - FoldedStr is represented as a string with the folded style.
    - LiteralStr is represented as a string with the literal style.
    - SingleQuotedStr is represented as a string with the single quoted style.
    - DoubleQuotedStr is represented as a string with the double quoted style.
    - FlowStyleSequence is represented as a list with the flow style.
    - FlowStyleMapping is represented as a dict with the flow style.

    These can be imported from drf_yaml.styles.
    """


SafeDumper.add_representer(
    decimal.Decimal,
    lambda dumper, data: dumper.represent_scalar(
        "tag:yaml.org,2002:str",
        force_str(data),
    ),
)

SafeDumper.add_representer(
    UUID,
    lambda dumper, data: dumper.represent_scalar(
        "tag:yaml.org,2002:str",
        force_str(data),
    ),
)

SafeDumper.add_representer(
    time,
    lambda dumper, data: dumper.represent_scalar(
        "tag:yaml.org,2002:str",
        force_str(data),
    ),
)

SafeDumper.add_representer(
    timedelta,
    lambda dumper, data: dumper.represent_float(
        data.total_seconds(),
    ),
)

SafeDumper.add_representer(
    bytes,
    lambda dumper, data: dumper.represent_scalar(
        "tag:yaml.org,2002:str",
        data.decode("utf-8"),
    ),
)

SafeDumper.add_representer(
    OrderedDict,
    yaml.representer.SafeRepresenter.represent_dict,
)

SafeDumper.add_representer(
    SafeString,
    yaml.representer.SafeRepresenter.represent_str,
)

SafeDumper.add_representer(
    Hyperlink,
    yaml.representer.SafeRepresenter.represent_str,
)

SafeDumper.add_representer(
    ErrorDetail,
    yaml.representer.SafeRepresenter.represent_str,
)

SafeDumper.add_representer(
    ReturnDict,
    yaml.representer.SafeRepresenter.represent_dict,
)

SafeDumper.add_representer(
    ReturnList,
    yaml.representer.SafeRepresenter.represent_list,
)

SafeDumper.add_representer(
    types.GeneratorType,
    yaml.representer.SafeRepresenter.represent_list,
)

SafeDumper.add_representer(
    styles.FoldedStr,
    lambda dumper, data: dumper.represent_scalar(
        "tag:yaml.org,2002:str",
        data,
        style=">",
    ),
)

SafeDumper.add_representer(
    styles.LiteralStr,
    lambda dumper, data: dumper.represent_scalar(
        "tag:yaml.org,2002:str",
        data,
        style="|",
    ),
)

SafeDumper.add_representer(
    styles.SingleQuotedStr,
    lambda dumper, data: dumper.represent_scalar(
        "tag:yaml.org,2002:str",
        data,
        style="'",
    ),
)

SafeDumper.add_representer(
    styles.DoubleQuotedStr,
    lambda dumper, data: dumper.represent_scalar(
        "tag:yaml.org,2002:str",
        data,
        style='"',
    ),
)

SafeDumper.add_representer(
    styles.FlowStyleSequence,
    lambda dumper, data: dumper.represent_sequence(
        "tag:yaml.org,2002:seq",
        data,
        flow_style=True,
    ),
)

SafeDumper.add_representer(
    styles.FlowStyleMapping,
    lambda dumper, data: dumper.represent_mapping(
        "tag:yaml.org,2002:map",
        data,
        flow_style=True,
    ),
)
